#include "CharsetImage.h"
CharsetImage::CharsetImage()
{
   PNGFile = NULL;
   charWidth = 0;
   charHeight = 0;
   wordWidth = 0;
   for(int i = 0; i < MAX_MAP_SIZE; i++) map[0] = 0;
   for(int i = 0; i < MAX_CHARSET_SIZE; i++) charSet[i] = NULL;
   png_ptr = NULL;
   info_ptr = NULL;
   png_pixels = NULL;
   row_pointers = NULL;
   chars = 0;
   mapTo = NULL;
   mapAs = NULL;
}

CharsetImage::~CharsetImage()
{
   for(int i = 0; i < MAX_CHARSET_SIZE; i++) if(charSet[i]) delete(charSet[i]);
   if(png_pixels) delete[](png_pixels);
   if(row_pointers) delete[](row_pointers);
   if(PNGFile) fclose(PNGFile);
   if(mapTo) delete(mapTo);
   if(mapAs) delete(mapAs);
}

CharsetImage*
CharsetImage::fromPNGFile(char* x)
{
   if(!this) return NULL;
   PNGFile = fopen(x, "rb");
   if(PNGFile == NULL)
   {
      printf("Can't open the file '%s'\n", x);
      return(NULL);
   }
   return this;
}

CharsetImage*
CharsetImage::withCharExtent(int width, int height)
{
   if(!this) return NULL;
   charWidth = width;
   charHeight = height;
   return this;
}

CharsetImage*
CharsetImage::withWordWidth(int width)
{
   if(!this) return NULL;
   wordWidth = width;
   return this;
}
CharsetImage*
CharsetImage::withMapping(unsigned char* m)
{
   if(!this) return NULL;
   if(m)
   {
      strncpy((char*)map, (char*)m, MAX_MAP_SIZE-1);
   }
   else
   {
      //create default mapping:
      //each pixel line occupies one or more words
      //and is aligned to the right
      if(charWidth == 0 || wordWidth == 0) return NULL;
      mapSizeInWords = (charWidth + wordWidth - 1) / wordWidth;
      //6 8 -> 11/8 = 1
      //7 8 -> 14/8 = 1
      //8 8 -> 15/8 = 1
      //9 8 -> 16/8 = 2 okay...
      map[charWidth * 2] = '\0';

      unsigned char w = 0;
      unsigned char wi = 'a' + mapSizeInWords - 1;
      for(int i = charWidth - 1; i >= 0; i--)
      {
         map[i * 2] = wi; //to word wi
         map[i * 2 + 1] = w + '0' + (w > 9) * ('A' - '9' - 1); //to bit w
         if(++w == wordWidth) 
         {
            w = 0;
            wi--;
         }
      }
      printf("Default mapping: %s\n", map);
   }
   return this;
}

CharsetImage*
CharsetImage::print(FILE* f)
{
   if(!this) return NULL;
   int n = widthInChars * heightInChars;
   for(int i = 0; i < n; i++)
   {
      unsigned char ch = (unsigned char)i + ' ';
      fprintf(f, "\tDW\t");
      charSet[i]->print(f);
      fprintf(f, "\t; %.1s\n", &ch);
   }
   return this;
}

CharsetImage*
CharsetImage::printWithStyle(FILE* f, char* style)
{
   if(!this) return NULL;
   int n = widthInChars * heightInChars;
   for(int i = 0; i < n; i++)
   {
      unsigned char ch = (unsigned char)i + ' ';
      fprintf(f, "\tDW ");
      charSet[i]->setStyle(style);
      charSet[i]->print(f);
      fprintf(f, "\t; %.1s\n", &ch);
   }
   return this;
}

CharsetImage*
CharsetImage::initialize()
{
   if(!this) return NULL;
   int ret;
   png_byte buf[8];
   int bit_depth;
   int color_type;
   png_uint_32 row_bytes;

   //read and check signature in PNG file
   ret = fread(buf, 1, 8, PNGFile);
   if(ret != 8) return NULL;
   ret = png_sig_cmp(buf, 0, 8);
   if(ret)
   {
      printf("Error: not a PNG file\n");
      return NULL;
   }

   //create png and info structures
   png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
      NULL, NULL, NULL);   //null pointers for error messages,
                           //error and warning functions.
   if(!png_ptr)
   {
      printf("Error: can't allocate memory for image data\n");
      return NULL;   //out of memory
   }

   info_ptr = png_create_info_struct(png_ptr);
   if(!info_ptr)
   {
      printf("Error: can't allocate memory for image data\n");
      png_destroy_read_struct(&png_ptr, NULL, NULL);
      return NULL;   //out of memory
   }

   //set jump back pointer
   if(setjmp(png_jmpbuf(png_ptr)))
   {
      printf("Error in libpng module\n");
      png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
      return NULL;
   }

   //set up the input control
   png_init_io(png_ptr, PNGFile);
   png_set_sig_bytes(png_ptr, 8);  //we already read the 8 signature bytes

   //read the file information
   png_read_info (png_ptr, info_ptr);

   //get size and bit-depth of the PNG-image
   width = png_get_image_width(png_ptr, info_ptr);
   height = png_get_image_height(png_ptr, info_ptr);
   if(width > MAX_IMAGE_WIDTH || height > MAX_IMAGE_HEIGHT)
   {
      printf("Error: Can't handle images larger than %dx%d pixels\n", MAX_IMAGE_WIDTH, MAX_IMAGE_HEIGHT);
      return NULL;
   }
   bit_depth = png_get_bit_depth(png_ptr, info_ptr);
   color_type = png_get_color_type(png_ptr, info_ptr);

   /* PNG_COLOR_TYPE_GRAY             PNG_COLOR_TYPE_RGB        
         (bit depths 1, 2, 4, 8, 16)      (bit_depths 8, 16)     
      PNG_COLOR_TYPE_GRAY_ALPHA        PNG_COLOR_TYPE_RGB_ALPHA  
         (bit depths 8, 16)               (bit_depths 8, 16)     
      PNG_COLOR_TYPE_PALETTE                                     
         (bit depths 1, 2, 4, 8)       PNG_COLOR_MASK_PALETTE    
      PNG_COLOR_MASK_COLOR             PNG_COLOR_MASK_ALPHA */   

   //set any format to gray
   png_set_expand(png_ptr); //expand to 24-bit RGB or 8-bit gray
   if(color_type != PNG_COLOR_TYPE_GRAY &&
      color_type != PNG_COLOR_TYPE_GRAY_ALPHA)
      png_set_rgb_to_gray(png_ptr, 1, -1, -1); //silent, and default for R and G

   //all transformations have been registered; now update info_ptr data,
   //get rowbytes and channels, and allocate image memory
   png_read_update_info (png_ptr, info_ptr);

   //row_bytes is the width x number of channels x (bit-depth / 8)
   row_bytes = png_get_rowbytes (png_ptr, info_ptr);
   png_pixels = new png_byte[row_bytes * height];
   row_pointers = new(png_byte*)[height];

   if(!png_pixels || !row_pointers)
   {
      png_destroy_read_struct (&png_ptr, &info_ptr, NULL);
      printf("Error: can't allocate memory for image data\n");
      return NULL;
   }

   // set the individual row_pointers to point at the correct offsets
   for(int i = 0; i < height; i++)
      row_pointers[i] = png_pixels + i * row_bytes;

   // now we can go ahead and just read the whole image
   png_read_image (png_ptr, row_pointers);

   // read rest of file
   png_read_end (png_ptr, NULL);  //we don't need any additional info

   // clean up after the read, and free any memory allocated
   png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

   //find image extent in characters
   if(charWidth == 0 || charHeight == 0)
   {
      printf("Error: character width and height is not specified");
      return NULL;
   }
   //create charset from the image
   if(!createCharSet()) return(NULL);
   return this;
}

CharsetImage*
CharsetImage::translateCharset()
{
   //check input data (if error return zero)
   //translate each character
   if(!chars) return(NULL);

   for(int i = 0; i < chars; i++)
      if(!charSet[i]->translate()) return NULL;

   return this;
}

CharsetImage*
CharsetImage::createCharSet()
{
   //Create charSet array of CharData and
   //initialize each CharData
   int i;
   int j;
   int row;
   int col;
   int line;

   //analyze mapping in the first place
   if(!analyzeMap()) return(NULL);

   //image extent in characters
   widthInChars = width / charWidth; //truncate fractional part
   heightInChars = height / charHeight; //truncate fractional part

   //total number of characters
   chars = widthInChars * heightInChars; 

   //scan all characters from left to right, from top to bottom
   for(i = 0, row = 0; row < heightInChars; row++)
   {
      for(col = 0; col < widthInChars; col++, i++)
      {
         if(!charSet[i])
         {
            //create new CharData and initialize it
            charSet[i] = new CharData;
            charSet[i]
               ->withCharExtent(charWidth, charHeight)
               ->withWordWidth(wordWidth)
               ->withMapping(mapTo, mapAs, mapSizeInWords, mapSizeInBits);
            line = row * charHeight * width + col * charWidth;
            for(j = 0; j < charHeight; j++)
            {
               charSet[i]->addPixelLine(png_pixels+line);
               line += width;
            }
         }
      }
   }

}

CharsetImage*
CharsetImage::analyzeMap()
{
   int n = strlen((char*)map);
   int i = 0;
   int j;
   int x;
   unsigned char z;
   unsigned char usedWords[MAX_WORDS_PER_CHAR];
   unsigned char to;
   unsigned long as;

   for(i = 0; i < MAX_WORDS_PER_CHAR; i++) usedWords[i] = 0;

   //check syntax
   for(x = 0; x < n; x += i)
   {
      for(i = 0; i < 2*charWidth && (x + i + 1) < n; i += 2)
      {
         //check first character
         if((map[x+i] >= 'a')
            && (map[x+i] < ('a'+MAX_WORDS_PER_CHAR)))
         {
            //remember the word was used
            int k = map[x+i] - 'a';
            if(k >= MAX_WORDS_PER_CHAR)
            {
               printf("Error in mapping: words number per character exceeded limit of %d\n", MAX_WORDS_PER_CHAR);
               return(NULL);
            }
            usedWords[k] = 1;
         }
         else if(map[x+i] == '-')
         {
            //this must be a start of a pixel line
            if(i != 0)
            {
               printf("Error in mapping: '-' in the wrong place\n");
               return NULL;
            }
            //correct indices to start from i = 0 and next character
            i = -2;
            x++;
            continue;
         }
         else
         {
            printf("Error in mapping: first character should be above or equal to 'a' - word index\n");
            return NULL;
         }
         //check second character
         z = map[x + i + 1];
         char max = wordWidth + (wordWidth >= 10) * ('A'-'9'-1) + '0' - 1;
         if(z < '0'
            || (z > '9' && z < 'A')
            || (z > max))
         {
            printf("Error in mapping: second character is a bit number (hex), can be 0..%.1s\n", &max);
            return NULL;
         }
      }
   }
   //syntax is okay, check number of used words
   mapSizeInWords = 0;
   for(i = 0; i < MAX_WORDS_PER_CHAR; i++)
   {
      if(!usedWords[i])
      {
         if(i == 0)
         {
            printf("Error in mapping: zero word (index 'a') not used\n");
            return NULL;
         }
         if(mapSizeInWords == 0)
         {
            mapSizeInWords = i;
         }
      }
      else
      {
         if(mapSizeInWords != 0)
         {
            z = i + 'a';
            printf("Error in mapping: unused word '%.1s' detected between the used ones\n", &z);
            return NULL;
         }
      }
   }

   //allocate memory for mapTo[] and mapAs[]
   mapTo = new unsigned char[mapSizeInWords * wordWidth];
   mapAs = new unsigned long[mapSizeInWords * wordWidth];
   if(mapTo == NULL || mapAs == NULL) return(NULL);

   //clear arrays
   for(i = 0; i < (mapSizeInWords * wordWidth); i++)
   {
      mapTo[i] = 0;
      mapAs[i] = 0;
   }
   //fill arrays
   j = 0;
   for(x = 0; x < n; x += i)
   {
      for(i = 0; i < 2*charWidth && (x + i + 1) < n; i += 2)
      {
         //check first character
         if(map[x+i] == '-')
         {
            //skip it and correct indices to start from i = 0 and next character
            i = -2;
            x++;
            continue;
         }
         else
         {
            //select word
            to = map[x+i] - 'a';
         }
         //check second character
         //convert ascii to decimal (case sensitive!)
         z = map[x + i + 1];
         if(z >= 'A')
         {
            z -= 'A'-('9'+1);
         }
         z -= '0';
         //convert to mask
         as = 1 << z;
         //write to arrays
         mapTo[j] = to;
         mapAs[j] = as;
         j++;
      }
   }
   mapSizeInBits = j;
   //done!
   return this;
}
