/* CharsetImage.h
 *
 * February 6, 2001 by Nikolai Golovchenko
 *
 * An object is declared for image of a charset and
 * a number of related functions (extraction and others)
 *
 * Description:
 * -----------
 * Image contains individual characters images
 * arranged from left to right and from top to down
 * without blank pixels between.
 *
 * A char is divided in horizontal lines of pixels.
 * Then each line is written into one or more
 * consecutive table entries. In some cases,
 * more than one line can fit in a word, and in some
 * others a line goes over several words.
 *
 * The user must decide which order of bits and what
 * locations to use. The mapping is specified by two characters
 * for each pixel in the line, from left to right. First character
 * tells which word to put the pixel in, and second character specifies
 * the bit in that word. For example, a0 means word number zero, bit 0
 * (least significant); bA - word 1, bit 10; cE - word 2, bit 14.
 * Unused bits in words are cleared. If more than one line mapping
 * is specified, they are separated by a dash.
 *
 * If the mapping wasn't specified, it is created by default. In that
 * case data is right aligned in the word (to the LSB) and no more than
 * one line is placed in a word.
 *
 * One character is placed in one group of words. Two characters data
 * cannot occupy one word.
 *
 * Examples;
 *   8x16 pixel characters, 8 bit word
 *      mapping=a7a6a5a4a3a2a1a0
 *   7x10 pixel characters, 8 bit word
 *      mapping=a6a5a4a3a2a1a0, (7th bit in entry unused and cleared)
 *      or      a7a6a5a4a3a2a1, (0th bit in entry unused and cleared)
 *      or      a3a2a1a0a6a5a4, etc  (7th bit in entry unused and cleared)
 *   12x16 pixel characters, 8 bit word
 *      mapping=a7a6a5a4a3a2a1a0b7b6b5b4-b3b2b1b0c7c6c5c4c3c2c1c0
 *         (two characters are packed in three words)
 *   5x7 pixel characters, 12 bit word
 *      mapping=a4a3a2a1a0-aBaAa9a8a7
 *         (two characters are packed in one word, bits 5 and 6 unused)
 *   10x7 pixel characters, 14 bit word (10+4.6+8.2+10+2.8+6.4+10)
 *      mapping=aDaCaBaAa9a8a7a6a5a4-a3a2a1a0bDbCbBbAb9b8-b7b6b5b4b3b2b1b0cDcC
 *             -cBcAc9c8c7c6c5c4c3c2-c1c0dDdCdBdAd9d8d7d6-...
 */

#ifndef CharImage_H
#define CharImage_H

#include "CharData.h"

class CharsetImage
{
public:
   //Initializing
   CharsetImage();
   ~CharsetImage();
   CharsetImage* fromPNGFile(char* x);
   CharsetImage* withCharExtent(int width, int height);
   CharsetImage* withWordWidth(int width);
   CharsetImage* withMapping(unsigned char* m);
   CharsetImage* initialize();
   CharsetImage* createCharSet();
   //Printing
   CharsetImage* print(FILE* f);
   CharsetImage* printWithStyle(FILE* f, char* style);
   //Accessing
   //Translating
   CharsetImage* analyzeMap();
   CharsetImage* translateCharset();
protected:
   int charWidth;
   int charHeight;   //from char sizes, the number of characters in a row
                     //and column is determined
   int wordWidth;    //how many bits in the target microcontroller table unit

   //mapping
   unsigned char map[MAX_MAP_SIZE]; //original form
   unsigned char* mapTo;            //translated to word index
   unsigned long* mapAs;            //and mask
   int mapSizeInWords;              //how many words the map covers
   int mapSizeInBits;               //how many character pixels in map

   //Image extent in characters
   int widthInChars;
   int heightInChars;
   int chars;

   //Translated characters data
   CharData* charSet[MAX_CHARSET_SIZE];

   //PNG image data
   FILE* PNGFile;       //image file pointer
   png_struct *png_ptr;
   png_info *info_ptr;
   png_uint_32 width;
   png_uint_32 height;
   png_byte** row_pointers;
   png_byte* png_pixels;
};

#endif
