#include <stdio.h>
#include <stdlib.h>

#include "png2charset.h"

void usage()
{
   printf("*******************************************************************************\n");
   printf("Charset extractor from a PNG image\n");
   printf("by Nikolai Golovchenko, 2001\n\n");
   printf("png2charset -w width -h height -b wordwidth [-m mapping] [-r hexprefix]\n\t in.png [out.asm]\n");
   printf("\n");
   printf("-w width       character extent in pixels\n");
   printf("-h height\n");
   printf("-b wordwidth   word width to translate the character data to\n");
   printf("-m mapping     optional mapping of each pixel line in a character\n");
   printf("               to words, like a7a6a5a4a3a2a1a0b7b6b5b4-c7c6c5c4c3c2c1c0b3b2b1b0\n");
   printf("-r hexprefix   prefix indicating a hex number ($ by default)\n");
   printf("*******************************************************************************\n");
}

int main(int argc, char *argv[])
{
   //options
   char* fout = NULL;
   char* fin = NULL;
   FILE* fileOut;
   int w = -1;
   int h = -1;
   int b = -1;
   unsigned char* m = NULL;
   char* style = "$";
   //charset image object
   CharsetImage* image;
   //misc
   int argi;

   if(argc == 1)
   {
      //no options
      usage();
      return(1);
   }

   //read options
   for (argi = 1; argi < argc; argi++)
   {
      if (argv[argi][0] == '-')
      {
         switch (argv[argi][1])
         {
         case 'w':
            w = atoi(argv[++argi]);
            break;
         case 'h':
            h = atoi(argv[++argi]);
            break;
         case 'b':
            b = atoi(argv[++argi]);
            break;
         case 'm':
            m = (unsigned char*)argv[++argi];
            break;
         case 'r':
            style = argv[++argi];
            break;
         case '?':
            usage();
            return(0);
         default:
            printf("Error: unknown option %s\n", argv[argi]);
            return(1);
         }
      }
      else
      {
         if(fin == NULL)
         {
            fin = argv[argi];
         }
         else if(fout == NULL)
         {
            fout = argv[argi];
         }
         else
         {
            printf("Error: too much parameters: %s\n", argv[argi]);
            return(1);
         }
      }
   }
   //check options
   if(w < 0 || h < 0)
   {
      printf("Error: character extent not specified, use options -w and -h\n");
      return(1);
   }
   if(w == 0 || h == 0)
   {
      printf("Error: wrong character size\n");
      return(1);
   }
   if(b == -1)
   {
      printf("Error: word width not specified, use -b option\n");
      return(1);
   }
   if(b == 0 || b > MAX_WORD_WIDTH)
   {
      printf("Error: wrong word width, should be 1..%d\n", MAX_WORD_WIDTH);
      return(1);
   }
   if(fin == NULL)
   {
      printf("Error: no input file\n");
      return(1);
   }
   if(fout == NULL)
   {
      fileOut = stdout;
   }
   else
   {
      fileOut = fopen(fout, "w");
   }
   if(fileOut == NULL)
   {
      printf("Error: can't create output file %s\n", fout);
      return(1);
   }
   if(m != NULL && strlen((char*)m) >= (MAX_MAP_SIZE - 1))
   {
      printf("Warning: map size truncated to %d characters\n", MAX_MAP_SIZE - 1);
   }



   image = new CharsetImage;
   if ((image->fromPNGFile(fin)
         ->withCharExtent(w, h)
         ->withWordWidth(b)
         ->withMapping(m)
         ->initialize())
         == NULL)
   {
      printf("Error while initializing\n");
      fclose(fileOut);
      delete(image);
      return(1);
   }
     

   if((image->translateCharset()) == NULL)
   {
      printf("Error while translating\n");
      fclose(fileOut);
      delete(image);
      return(1);
   }

   image->printWithStyle(fileOut, style);
   fclose(fileOut);
   delete(image);
   return(0);
}

